import logging

import cvxpy as cp
import numpy as np

logger = logging.getLogger(__name__)


class GeneralizedDivergenceSolver(object):
    def fit(self, X, mu_multisample):
        """TODO docstring"""
        self.opt_res_list = []
        n = mu_multisample.shape[0]
        for i in range(n):
            logger.info(f"Solving for sample i={i}")

            opt_res = self._fit1sample(X, mu_multisample[i, :])

            self.opt_res_list.append(opt_res)

    # TODO: Add option to incorporate size_factors?
    def weights(self, renormalize=True):
        if not self.opt_res_list:
            raise ValueError("Need to call fit() first")

        w_hat_multisample = []
        for i in range(len(self.opt_res_list)):
            w_hat_multisample.append(self._weights1sample(i))

        ret = np.array(w_hat_multisample).T
        if renormalize:
            ret[ret < 0] = 0
            ret = np.einsum("ij,j->ij", ret, 1.0 / ret.sum(axis=0))

        return ret

    @property
    def _equality_constraints(self):
        return (self.mom_atol == 0) and (self.mom_rtol == 0)


class AlphaDivergenceCvxpySolver(GeneralizedDivergenceSolver):
    def __init__(
        self, alpha, mom_atol=0, mom_rtol=0, solve_kwargs=None, use_norm=False
    ):
        """
        :param float alpha: Value of alpha for alpha-divergence.
            Also accepts 'pearson' for alpha=2 (which is a quadratic program)
            or 'kl' for alpha=1 (which is same as maximum entropy).
        :param float mom_atol: For moment constraints, require
            X.T @ w = mu +/- eps, where
            eps = mom_atol + abs(mu) * mom_rtol.
        :param float mom_rtol: For moment constraints, require
            X.T @ w = mu +/- eps, where
            eps = mom_atol + abs(mu) * mom_rtol.
        :param dict solve_kwargs: Additional kwargs to pass to
            `cvxpy.Problem.solve`.
        :param bool use_norm: Replaces the f-divergence with a monotone
            transformation that may have better conditioning. Specifically,
            optimizes sum((n*w)**alpha)**(1/alpha) - n**(1/alpha).
            Note it prevents using efficient QP solvers when alpha=2.
            Requires alpha>1.
        """
        if alpha == "pearson":
            alpha = 2
        elif alpha == "kl":
            alpha = 1
        elif isinstance(alpha, str):
            raise ValueError(f"Unrecognized divergence {alpha}")

        self.alpha = alpha
        self.mom_atol = mom_atol
        self.mom_rtol = mom_rtol

        # Reference for solver options:
        # https://www.cvxpy.org/tutorial/advanced/index.html#setting-solver-options
        if not solve_kwargs:
            solve_kwargs = {}
        solve_kwargs.setdefault("verbose", False)
        self.solve_kwargs = solve_kwargs

        if use_norm:
            if alpha <= 1:
                raise NotImplementedError("use_norm requires alpha > 1")
        self.use_norm = use_norm

        self.opt_res_list = []

    def dual_nonneg(self):
        # FIXME Don't make the index manually hardcoded, to prevent
        # accidental future breakage
        return np.array([prob.constraints[0].dual_value for prob in self.opt_res_list])

    def dual_sum1(self):
        # FIXME Don't make the index manually hardcoded, to prevent
        # accidental future breakage
        return np.array([prob.constraints[1].dual_value for prob in self.opt_res_list])

    # HACK Hardcoding that the first 2 constraints are the sum-to-1
    # and non-negative constraints. Assert this when constructing
    # constraints.
    @property
    def _dual_n_skip_constraints(self):
        return 2

    def _dual_moments(self):
        start = self._dual_n_skip_constraints
        if self._equality_constraints:
            vals = []
            for i in range(len(self.opt_res_list)):
                vals.append(self.opt_res_list[i].constraints[start].dual_value)
            vals = np.array(vals)
            return {"dual": vals}
        else:
            # FIXME Don't make the fact that upper is first
            # manually hardcoded
            upper = np.array(
                [
                    self.opt_res_list[i].constraints[start].dual_value
                    for i in range(len(self.opt_res_list))
                ]
            )
            # FIXME Don't make the fact that lower is second
            # manually hardcoded
            lower = np.array(
                [
                    self.opt_res_list[i].constraints[start + 1].dual_value
                    for i in range(len(self.opt_res_list))
                ]
            )
            return {"dual": upper - lower, "dual_upper": upper, "dual_lower": lower}

    def _objective(self):
        return np.array([prob.value for prob in self.opt_res_list])

    def _weights1sample(self, i):
        """TODO docstring"""
        (w_hat,) = self.opt_res_list[i].variables()
        return w_hat.value.copy()

    # TODO: Add accelerated projected gradient descent solver? E.g. see:
    # https://stackoverflow.com/questions/65526377/cvxpy-returns-infeasible-inaccurate-on-quadratic-programming-optimization-proble
    def _fit1sample(self, X, mu):
        """Estimate density weights for a single sample on a single-cell reference.

        :param `numpy.ndarray` X: Reference embedding. Rows=cells, columns=features.
        :param `numpy.ndarray` mu: Sample moments. Either bulk gene counts (for bulk
            deconvolution) or sample centroids of single cells (for differential
            abundance). Should be a 1-dimensional array.

        :rtype: :class:`cvxpy.Problem`
        """

        n = X.shape[0]
        z = np.zeros(n)

        w = cp.Variable(n)

        # Initialize as uniform distribution
        # NOTE: Unclear in which solvers CVXPY will actually use the initialization
        # https://www.cvxpy.org/tutorial/advanced/index.html#warm-start
        # https://stackoverflow.com/questions/52314581/initial-guess-warm-start-in-cvxpy-give-a-hint-of-the-solution
        w.value = np.ones(n) / n

        Xt = X.T

        if self.use_norm:
            assert self.alpha > 1
            # Minimum at 0 when w is uniform, by Holder's inequality
            objective = n * cp.norm(w, self.alpha) - n ** (1 / self.alpha)
            # NOTE: This claims use_norm is preferable even when alpha=2:
            # http://cvxr.com/cvx/doc/advanced.html#eliminating-quadratic-forms
            # But I'm dubious about it -- I think it holds for conic
            # solvers like ECOS, but not dedicated QP solvers like OSQP
        elif self.alpha == 1:
            objective = -cp.sum(cp.entr(w)) - np.log(n)
        elif self.alpha == 0:
            objective = -cp.sum(cp.log(w)) - n * np.log(n)
        else:
            objective = (
                (n ** (self.alpha - 1) * cp.sum(w**self.alpha) - 1)
                / self.alpha
                / (self.alpha - 1)
            )

        objective = cp.Minimize(objective)

        constraints = [w >= z, cp.sum(w) == 1.0]

        # NOTE Hardcoding that the first 2 constraints are the
        # non-interesting ones. dual_moments checks this when figuring
        # which constraint to use
        assert len(constraints) == self._dual_n_skip_constraints

        if self._equality_constraints:
            constraints.append(
                Xt @ w == mu,
            )
        else:
            eps = self.mom_atol + self.mom_rtol * np.abs(mu)

            # NOTE dual_moments assumes the upper constraint is always
            # first. FIXME Make the ordering of upper/lower more
            # robust/automatic instead of manually hardcoded. Or at
            # least add some kind of assertion here!

            constraints.append(Xt @ w - mu <= eps)

            constraints.append(Xt @ w - mu >= -eps)

        prob = cp.Problem(objective, constraints)

        res = prob.solve(**self.solve_kwargs)
        assert prob.variables()[0] is w
        assert prob.value is res

        # TODO: Reenable assertions, with params for atol/rtol?
        # w_hat, = prob.variables()
        # w_hat = w_hat.value
        # assert np.all(np.isclose(w_hat, 0, atol=1e-4) | (w_hat >= 0))
        # assert np.allclose(np.sum(w_hat), 1)

        logger.info(f"objective={prob.value}, {prob.status}")

        return prob


# TODO: Implement Vajda's chi-alpha divergence?
# See the table on the f-divergence wikipedia page.
# It's equivalent to the Lp-norm of w minus uniform.



import logging

import jax
import jax.numpy as jnp
import scipy
import scipy.optimize

from ._solvers import GeneralizedDivergenceSolver

logger = logging.getLogger(__name__)


def maxent_dual(lambd, b, A):
    return -b @ lambd - jnp.log(jnp.sum(jnp.exp(-A.T @ lambd)))


def maxent_solution(lambd_star, A):
    nu_star = jnp.log(jnp.sum(jnp.exp(-A.T @ lambd_star))) - 1
    return 1.0 / jnp.exp(A.T @ lambd_star + nu_star + 1)


def maxent_solve_dual_lbfgs(A, b, n_inequality, opt_kwargs=None):
    """Solve maximum entropy via the dual and L-BFGS-B. Based on Boyd
    & Vandenberge's Convex Optimization, Example 5.3, with the same
    notation. A minor difference difference is we allow equality
    constraints; the first n_inequality rows of A correspond to
    inequalities (a_i^T x <= b) while the remaining rows correspond to
    equalities (a_i^T x = b). opt_kwargs are passed to
    scipy.optimize.minimize. Returns a dictionary with the primal
    solution as well as the optimizer output for the dual.
    """

    def optfun(lambd):
        return -maxent_dual(lambd, b, A)

    gradfun = jax.grad(optfun)

    bounds = []
    for _ in range(n_inequality):
        bounds.append((0, None))
    for _ in range(A.shape[0] - n_inequality):
        bounds.append((None, None))

    if not opt_kwargs:
        opt_kwargs = {}

    res = scipy.optimize.minimize(
        optfun,
        jnp.zeros(A.shape[0]),
        jac=gradfun,
        bounds=bounds,
        method="L-BFGS-B",
        **opt_kwargs,
    )

    return {"primal": maxent_solution(res.x, A), "dual_opt_res": res}


class MaxentDualLbfgsSolver(GeneralizedDivergenceSolver):
    def __init__(self, mom_atol=0, mom_rtol=0, solve_kwargs=None):
        self.mom_atol = mom_atol
        self.mom_rtol = mom_rtol

        if not solve_kwargs:
            solve_kwargs = {}
        self.solve_kwargs = solve_kwargs

        self.opt_res_list = []

    def _weights1sample(self, i):
        return self.opt_res_list[i]["primal"]

    def _fit1sample(self, X, mu):
        if self.mom_atol == 0 and self.mom_rtol == 0:
            A = -X.T
            b = -mu
            n_inequality = 0
        else:
            eps = self.mom_atol + self.mom_rtol * jnp.abs(mu)
            A = jnp.vstack([X.T, -X.T])
            b = jnp.concatenate([mu + eps, -mu + eps])
            n_inequality = A.shape[0]

        res = maxent_solve_dual_lbfgs(A, b, n_inequality, opt_kwargs=self.solve_kwargs)

        logger.info(
            f"objective={res['dual_opt_res']['fun']}, {res['dual_opt_res']['message']}"
        )

        return res

    def _objective(self):
        w = self.weights(renormalize=False)
        return jnp.sum(w * jnp.log(w), axis=0) - jnp.log(w.shape[0])

    def _dual_moments(self):
        vals = []
        for i in range(len(self.opt_res_list)):
            vals.append(self.opt_res_list[i]["dual_opt_res"].x)
        vals = jnp.array(vals)

        if self._equality_constraints:
            # Note we use -X, -mu in the equality case
            return {"dual": -vals}
        else:
            nvals = vals.shape[1] / 2
            assert nvals == round(nvals)
            nvals = int(nvals)
            # FIXME Don't make the fact that upper is the first half
            # manually hardcoded?
            upper = vals[:, :nvals]
            lower = vals[:, nvals:]
            return {"dual": upper - lower, "dual_upper": upper, "dual_lower": lower}



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.1
#   kernelspec:
#     display_name: quipcell
#     language: python
#     name: quipcell
# ---

# %% [markdown]
# # Example application of quipcell to human lung cell atlas
#
# In this vignette, we demonstrate the application of `quipcell` to bulk deconvolution, replicating some analyses from the manuscript.
#
# For our dataset, we use the Human Lung Cell Atlas, which consists of several studies. We hold out one of the studies as a test set, which we use to generate (pseudo)bulk samples with known ground truth.
#
# To make this analysis runnable on a laptop, the data has been preprocessed into a smaller form (just the raw counts of the highly variable genes). See the [vignette readme](https://github.com/Genentech/quipcell/blob/main/vignette/Readme.md) for instructions on how to download or generate the preprocessed data before running this notebook.

# %%
from collections import Counter

import logging
logging.basicConfig(level=logging.INFO)

import numpy as np
import pandas as pd

import plotnine as gg
import mizani
import mizani.transforms as scales

import anndata as ad
import scanpy as sc

import scipy.sparse

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.preprocessing import OneHotEncoder

import quipcell as qpc

# %% [markdown]
# ## Load single cell data
#
# First, we load the (preprocessed) Human Lung Cell Atlas data.

# %%
adata = ad.read_h5ad("data/hlca_hvgs.h5ad")


# %% [markdown]
# Next, we subset to 1/8th of the dataset to improve speed. We also exclude one of the studies as a validation test set to benchmark bulk deconvolution.

# %%
validation_study = 'Krasnow_2020'

rng = np.random.default_rng(12345)
n = adata.obs.shape[0]

keep = adata.obs['study'] != validation_study
keep = keep & (rng.random(size=n) < .125)

adata_ref = adata[keep,:]
print(f"{adata_ref.shape[0]} cells in reference set")

# %% [markdown]
# ## Normalize and transform single cell data
#
# Next, normalize the raw counts. The preprocessing script saved the total UMIs per cell before filtering to highly variable genes, so we divide it out, and normalize to counts per 1000 UMIs.

# %%
# Normalize
normalize_sum = 1000
adata_ref.X = scipy.sparse.diags(normalize_sum / adata_ref.obs['n_umi'].values) @ adata_ref.X

# %% [markdown]
# Next, we run PCA, and then perform Linear Discriminant Analysis on the PCs and celltype labels. This will be the embedding space we use for `quipcell`.

# %%
sc.pp.pca(adata_ref, n_comps=100)

# since the PCA is zero-centered, we need to save the gene-wise means
# to apply PCA rotation to pseudobulks
gene_means = adata_ref.X.mean(axis=0)
gene_means = np.squeeze(np.asarray(gene_means))
adata_ref.var['x_mean'] = gene_means

lda = LinearDiscriminantAnalysis(n_components=15)

X = adata_ref.obsm['X_pca']
y = adata_ref.obs['ann_finest_level']

lda.fit(X, y)

adata_ref.obsm['X_lda'] = lda.transform(adata_ref.obsm['X_pca'])

# %% [markdown]
# ## Load and transform (pseudo)bulk data
#
# Next, we read in the pseudobulk counts (generated in `preprocessing.py`), and subset to the samples in the validation/test set.
#
# We then normalize the counts, and subset to the same genes as the reference single cell set.
#
# Then, we apply a linear transformation to map the bulk samples to the embedding space.

# %%
adata_pseudobulk = ad.read_h5ad("data/pseudobulks.h5ad")

# Subset to validation set
keep = adata_pseudobulk.obs['study'] == validation_study
adata_pseudobulk = adata_pseudobulk[keep,:]

# Normalize on the same scale as the single cell reference set
adata_pseudobulk.raw = adata_pseudobulk
sc.pp.normalize_total(adata_pseudobulk, target_sum=normalize_sum)

# Subset to same genes as the single cell reference set
adata_pseudobulk = adata_pseudobulk[:, adata.var.index]

# Apply PCA rotation to pseudobulks
X = adata_pseudobulk.X - adata_ref.var['x_mean'].values
X = np.asarray(X @ adata_ref.varm['PCs'])
adata_pseudobulk.obsm['X_pca'] = X
# Apply LDA rotation to pseudobulks
adata_pseudobulk.obsm['X_lda'] = lda.transform(X)

# %% [markdown]
# ## Estimate weights for read-level abundances
#
# Now that the single cell reference set and the bulk samples are in the same embedding space, we can estimate the weights with `quipcell`.

# %%
w_umi = qpc.estimate_weights_multisample(adata_ref.obsm['X_lda'],
                                         adata_pseudobulk.obsm['X_lda'])

# %% [markdown]
# ## Convert read-level abundance to cell-level abundance
#
# Note that the gene counts of bulk samples don't normalize for the fact that different cell types have different number of reads. Therefore, the weights represent a probability distribution of a random _read_ drawn from the bulk sample, rather than a probability distribution over a random _cell_ from the bulk sample.
#
# We can convert the read-level weights to cell-level weights by estimating size factors for each of the reference cells, and normalizing the weights by those.
#
# If the reference set doesn't have too much technical variation, then we can just use the number of UMIs per cell as the normalizing factor. However, the HLCA is a highly heterogeneous dataset, consisting of many samples from different studies, sequenced to different depths, and containing different celltypes. Therefore, we need to control for sample when estimating the size factors. `quipcell` provides a method to estimate size factors while controlling for sample by using Poisson regression.

# %%
# Estimate size factors using Poisson regression
size_factors = qpc.estimate_size_factors(
    adata_ref.obsm['X_lda'],
    adata_ref.obs['n_umi'].values,
    adata_ref.obs['sample'].values,
    #verbose=True
)

# Convert read-level weights to cell-level weights
w_cell = qpc.renormalize_weights(w_umi, size_factors)

# %% [markdown]
# ## Plot inferred weights
#
# Below, we plot the read-level and cell-level weights on the reference UMAP for each of the 5 validation bulk samples.

# %%
# Dataframe for plotting UMI-level weights on UMAP
df = pd.DataFrame(
    np.hstack([adata_ref.obsm['X_umap'], w_umi]),
    columns = ['UMAP1', 'UMAP2'] + list(adata_pseudobulk.obs.index)
).melt( # pivot to long format
    id_vars=['UMAP1', 'UMAP2'],  
    var_name='sample', 
    value_name='weight'
)

# set very small weights to 0 for prettier plotting
weight_trunc = df['weight'].values.copy()
weight_trunc[weight_trunc < 1e-9] = 0
df['weight_trunc'] = weight_trunc

(gg.ggplot(df, gg.aes(x="UMAP1", y="UMAP2", color="weight_trunc")) +
    gg.geom_point(size=.25, alpha=.5) +
    gg.facet_wrap("~sample") +
    gg.scale_color_cmap(trans=scales.log_trans(base=10)) +
    gg.ggtitle("Read-level weights"))

 # %%
 # Dataframe for plotting cell-level weights on UMAP
df = pd.DataFrame(
    np.hstack([adata_ref.obsm['X_umap'], w_cell]),
    columns = ['UMAP1', 'UMAP2'] + list(adata_pseudobulk.obs.index)
).melt( # pivot to long format
    id_vars=['UMAP1', 'UMAP2'],  
    var_name='sample', 
    value_name='weight'
)

# set very small weights to 0 for prettier plotting
weight_trunc = df['weight'].values.copy()
weight_trunc[weight_trunc < 1e-9] = 0
df['weight_trunc'] = weight_trunc

(gg.ggplot(df, gg.aes(x="UMAP1", y="UMAP2", color="weight_trunc")) +
    gg.geom_point(size=.25, alpha=.5) +
    gg.facet_wrap("~sample") +
    gg.scale_color_cmap(trans=scales.log_trans(base=10)) +
    gg.ggtitle("Cell-level weights"))

# %% [markdown]
# ## Plot true vs estimated abundances
#
# Next, we compare the estimated weights to the true celltype proportions. In the preprocessing script, we precomputed the fraction of reads and cells coming from each celltype, at different levels of celltype annotation. We read in these precomputed abundances below.

# %%
df_abundance = pd.read_csv('data/abundances.csv')
# filter to samples in the validation set
df_abundance = df_abundance[df_abundance['sample'].isin(adata_pseudobulk.obs.index)]
# convert sample name to string to prevent missing categorical levels
df_abundance['sample'] = df_abundance['sample'].astype(str)
df_abundance

# %% [markdown]
# Next, we compute estimated celltype abundances, by summing over the weights fit by `quipcell`. We add the estimated number of reads and cells coming from each celltype to the dataframe of abundances.

# %%
est_frac_umi = []
est_frac_cells = []

# TODO use a more descriptive variable name than i
for i in range(1, 6):
    enc = OneHotEncoder()
    mat_onehot = enc.fit_transform(
        adata_ref.obs[f'celltype_lvl{i}'].values.reshape(-1,1)
    )

    # Function to sum weights by celltype, then pivot to long dataframe
    def aggregate_and_reshape(w):
        return pd.DataFrame(
            w.T @ mat_onehot,
            columns = enc.categories_[0],
            index = adata_pseudobulk.obs.index
        ).reset_index(names='sample').melt(
            id_vars=['sample'], var_name='celltype', value_name='weight'
        ).assign(ann_level=f'lvl{i}')

    est_frac_umi.append(aggregate_and_reshape(w_umi))
    est_frac_cells.append(aggregate_and_reshape(w_cell))


# Function to combine the summed weights across annotation levels,
# returning a vector to be added to df_abundance
def concat_aggregated_weights(agg_weight_list):
    idx_keys = ['ann_level', 'sample', 'celltype']
    return (pd.concat(agg_weight_list)
            .set_index(idx_keys)
            .loc[zip(*[df_abundance[k] for k in idx_keys])]
            .values)

df_abundance['est_frac_umi'] = concat_aggregated_weights(est_frac_umi)
df_abundance['est_frac_cell'] = concat_aggregated_weights(est_frac_cells)

df_abundance

# %% [markdown]
# Finally, we plot the estimated vs actual abundances, at both the read and cell level.
#
# At the finest resolution (annotation level 5), there is some inaccuracy because a common celltype in the validation set (CCL3+ Alveolar macrophages) is very rare in the reference samples, and cannot be distinguished from other Alveolar macrophages in the embedding space. See the manuscript for more details.

 # %%
 (gg.ggplot(df_abundance.sample(frac=1, random_state=42), 
           gg.aes(x="frac_umi", y="est_frac_umi", color="ann_level", shape="sample")) +
    gg.geom_point() +
    gg.geom_abline(linetype='dashed') +
    gg.scale_x_sqrt(breaks=[.01,.1,.2,.4,.6,.8]) +
    gg.scale_y_sqrt(breaks=[.01,.1,.2,.4,.6,.8]) +
    gg.theme_bw(base_size=16) +
    gg.ggtitle("Read-level accuracy"))

 # %%
 (gg.ggplot(df_abundance.sample(frac=1, random_state=42), 
           gg.aes(x="frac_cell", y="est_frac_cell", color="ann_level", shape="sample")) +
    gg.geom_point() +
    gg.geom_abline(linetype='dashed') +
    gg.scale_x_sqrt(breaks=[.01,.1,.2,.4,.6,.8]) +
    gg.scale_y_sqrt(breaks=[.01,.1,.2,.4,.6,.8]) +
    gg.theme_bw(base_size=16) +
    gg.ggtitle("Cell-level accuracy"))

# %%



import logging

import numpy as np
from sklearn import linear_model
from sklearn.preprocessing import OneHotEncoder

from ._maxent_dual_lbfgs import MaxentDualLbfgsSolver
from ._solvers import AlphaDivergenceCvxpySolver

logger = logging.getLogger(__name__)


# TODO: parametrize the cvxpy.Problem by mu to reduce compilation times?
# https://www.cvxpy.org/tutorial/advanced/index.html#disciplined-parametrized-programming
def estimate_weights_multisample(
    X,
    mu_multisample,
    renormalize=True,
    alpha="pearson",
    use_dual_lbfgs=None,
    return_with_solver_info=False,
    mom_atol=0,
    mom_rtol=0,
    solve_kwargs=None,
):
    """Estimate density weights for multiple samples on a single-cell reference.

    Returns a numpy array with rows=cells, columns=samples.

    :param `numpy.ndarray` X: Reference embedding. Rows=cells, columns=features.
    :param `numpy.ndarray` mu_multisample: Sample moments. Either bulk gene counts
        or a linear transformation thereof. Rows=samples, columns=features.
    :param bool renormalize: Correct for any solver inaccuracies by setting small
        negative numbers to 0, and renormalizing sums to 1.
    :param float alpha: Value of alpha for alpha-divergence. Also accepts 'pearson'
        for alpha=2 (which is a quadratic program) or 'kl' for alpha=1
        (which is same as maximum entropy).
    :param bool use_dual_lbfgs: If True, solve via the dual problem with L-BFGS-B,
        instead of using cvxpy. Currently only implemented for alpha==1 or
        alpha=='kl'. Default is True when alpha==1 or 'kl', and False otherwise.
    :param float mom_atol: For moment constraints, require
        X.T @ w = mu +/- eps, where
        eps = mom_atol + abs(mu) * mom_rtol.
    :param float mom_rtol: For moment constraints, require
        X.T @ w = mu +/- eps, where
        eps = mom_atol + abs(mu) * mom_rtol.
    :param dict solve_kwargs: Additional kwargs to pass to
        :meth:`cvxpy.Problem.solve` or :meth:`scipy.optimize.minimize` (depending
        on ``use_dual_lbfgs``).
    :param bool return_with_solver_info: If False, return a numpy array with the
        weights.  If True, return a dict with the weights, the dual solution for
        the moment constraints, and the objective.

    :rtype: Union[dict,numpy.ndarray]
    """
    if use_dual_lbfgs is None:
        use_dual_lbfgs = alpha == 1 or alpha == "kl"

    if not use_dual_lbfgs:
        solver = AlphaDivergenceCvxpySolver(
            alpha=alpha, mom_atol=mom_atol, mom_rtol=mom_rtol, solve_kwargs=solve_kwargs
        )
    elif alpha == 1 or alpha == "kl":
        solver = MaxentDualLbfgsSolver(
            mom_atol=mom_atol, mom_rtol=mom_rtol, solve_kwargs=solve_kwargs
        )
    else:
        raise NotImplementedError("Dual LBFGS solver only implemented for alpha=1")

    solver.fit(X, mu_multisample)
    ret = solver.weights(renormalize=renormalize)
    if not return_with_solver_info:
        return ret
    else:
        ret = {"weights": ret, "objective": solver._objective()}
        ret.update(solver._dual_moments())
        return ret


def renormalize_weights(weights, size_factors):
    """Renormalize weights to correct for cell-specific size factors.

    This can be used to convert read-level probability weights to
    cell-level weights.

    :param `numpy.ndarray` weights: Weights as returned by
        `quipcell.estimate_weights_multisample`. Rows=cells, columns=samples.
    :param `numpy.ndarray` size_factors: Cell-level size factors. Should be
        1 dimensional, with length equal to the number of rows in `weights`.

    :rtype: :class:`numpy.ndarray`
    """
    weights = np.einsum("ij,i->ij", weights, 1 / size_factors)
    weights = np.einsum("ij,j->ij", weights, 1 / weights.sum(axis=0))
    assert np.allclose(weights.sum(axis=0), 1)
    return weights


def estimate_size_factors(X, n_reads, sample, **kwargs):
    """Estimate cell-level size factors with Poisson regression to control for
    technical variation between samples.

    Returns a vector of the size factor for each cell.

    ``**kwargs`` are additional keyword arguments passed to
        :class:`sklearn.linear_model.PoissonRegressor`.

    :param `numpy.ndarray` X: Single cell embedding. Rows=cells, columns=features.
    :param `numpy.ndarray` n_reads: 1d vector of the number of reads per cell.
    :param `numpy.ndarray` sample: 1d string vector for the sample that each cell
        came from.

    :rtype: :class:`numpy.ndarray`
    """
    enc = OneHotEncoder()

    modmat = enc.fit_transform(sample.reshape(-1, 1))
    modmat = np.hstack([X, np.asarray(modmat.todense())])

    kwargs.setdefault("fit_intercept", False)
    kwargs.setdefault("alpha", 0)
    kwargs.setdefault("solver", "newton-cholesky")
    kwargs.setdefault("verbose", 0)

    clf = linear_model.PoissonRegressor(**kwargs)

    clf.fit(modmat, n_reads)

    size_factors = np.exp(X @ clf.coef_[: X.shape[1]])
    size_factors = size_factors / np.mean(size_factors)
    return size_factors



#!/usr/bin/env python

import numpy as np
import pandas as pd

import anndata as ad
import scanpy as sc

import logging
logging.basicConfig(level=logging.INFO,
                    format="%(asctime)s %(message)s")

logging.info("Loading HLCA")

# Takes 17-18 GB RAM
adata = ad.read_h5ad("data/hlca_orig.h5ad")

# Generate & save pseudobulks
logging.info("Generating pseudobulks")

onehot = pd.get_dummies(adata.obs['sample'], dtype=float, sparse=True)
onehot_mat = onehot.sparse.to_coo().tocsr()

pseudobulks = onehot_mat.T @ adata.raw.X

obs_pseudobulk = (adata.obs[['sample', 'donor_id', 'study']]
      .drop_duplicates()
      .reset_index(drop=True))

assert set(obs_pseudobulk['sample']) == set(onehot.columns)

obs_pseudobulk['sample'] = obs_pseudobulk['sample'].astype(str)
obs_pseudobulk = obs_pseudobulk.set_index('sample')
obs_pseudobulk = obs_pseudobulk.loc[onehot.columns]

adata_pseudobulk = ad.AnnData(
    X=pseudobulks,
    obs=obs_pseudobulk,
    var=adata.var
)

logging.info("Saving pseudobulks")
adata_pseudobulk.write_h5ad("data/pseudobulks.h5ad")

# Save number of UMIs before renormalizing/subsetting
adata.obs['n_umi'] = adata.raw.X.sum(axis=1)

# Select highly variable genes
logging.info("Selecting highly variable genes")

sc.pp.highly_variable_genes(
    adata,
    batch_key='study'
)

# Reduce memory usage to enable analysis on laptop: Delete log counts,
# and filter to highly variable genes only
adata.X = adata.raw.X
del adata.raw

adata = adata[:,adata.var['highly_variable']]

# Adjust celltype annotations; at each level, replace "None" with the
# next level up
logging.info("Adjust celltype annotations (removing Nones)")

adata.obs['celltype_lvl1'] = adata.obs['ann_level_1']

for i in range(2,6):
    ann_finest = adata.obs[f'ann_level_{i}'].astype(str)
    is_none = (ann_finest == 'None')
    ann_finest[is_none] = adata.obs[f'celltype_lvl{i-1}'][is_none]
    adata.obs[f'celltype_lvl{i}'] = ann_finest

assert (adata.obs['ann_finest_level'] == adata.obs['celltype_lvl5']).all()

# save the modified anndata
logging.info("Saving modified anndata")
adata.write_h5ad("data/hlca_hvgs.h5ad")

# Make a dataframe of number of UMIs, cells from each celltype
logging.info("Make dataframe of celltype abundances")

df_abundance = []

for i in range(1, 6):
    df = adata.obs.rename(columns={f'celltype_lvl{i}': 'celltype'})

    df_agg = (df[['sample', 'celltype', 'n_umi']]
    .groupby(['sample', 'celltype'], observed=False)
    .sum()
    .reset_index())

    df_agg['n_umi'] = df_agg['n_umi'].astype(int)

    umi_denom = (df_agg[['sample', 'n_umi']]
             .groupby('sample', observed=False)
             .sum()
             .loc[df_agg['sample']]['n_umi']
             .values)

    df_agg['frac_umi'] = df_agg['n_umi'] / umi_denom

    n_cells = (df[['sample', 'celltype']]
    .groupby(['sample', 'celltype'], observed=False)
    .size())

    df_agg['n_cell'] = (n_cells
                        .loc[zip(df_agg['sample'], df_agg['celltype'])]
                        .values)

    cell_denom = (df_agg[['sample', 'n_cell']]
             .groupby('sample', observed=False)
             .sum()
             .loc[df_agg['sample']]['n_cell']
             .values)

    df_agg['frac_cell'] = df_agg['n_cell'] / cell_denom

    df_agg.insert(1, 'ann_level', f'lvl{i}')
    df_abundance.append(df_agg)

df_abundance = pd.concat(df_abundance)

df_abundance.to_csv("data/abundances.csv", index=False)



import os

import cvxpy as cp
import numpy as np

import quipcell as qpc

dirname = os.path.dirname(os.path.realpath(__file__))


def test_example():
    w = np.loadtxt(os.path.join(dirname, "test_example_w.txt"))

    x = np.loadtxt(os.path.join(dirname, "test_example_x.txt"))

    mu = np.loadtxt(os.path.join(dirname, "test_example_mu.txt"))

    w2 = qpc.estimate_weights_multisample(x, mu, solve_kwargs={"solver": cp.OSQP})

    assert np.allclose(w, w2)


def test_example_relaxed():
    w = np.loadtxt(os.path.join(dirname, "test_example_w_relaxed.txt"))

    x = np.loadtxt(os.path.join(dirname, "test_example_x.txt"))

    mu = np.loadtxt(os.path.join(dirname, "test_example_mu.txt"))

    w2 = qpc.estimate_weights_multisample(
        x, mu, mom_atol=0.001, solve_kwargs={"solver": cp.OSQP}
    )

    assert np.allclose(w, w2)


def test_example_relaxed2():
    w = np.loadtxt(os.path.join(dirname, "test_example_w_relaxed2.txt"))

    x = np.loadtxt(os.path.join(dirname, "test_example_x.txt"))

    mu = np.loadtxt(os.path.join(dirname, "test_example_mu.txt"))

    w2 = qpc.estimate_weights_multisample(
        x, mu, mom_rtol=0.1, solve_kwargs={"solver": cp.OSQP}
    )

    assert np.allclose(w, w2)


def test_dual_small_epsilon():
    x = np.loadtxt(os.path.join(dirname, "test_example_x.txt"))

    mu = np.loadtxt(os.path.join(dirname, "test_example_mu.txt"))

    # FIXME switch to:
    # estimate_weights_multisample(return_with_solver_info=True)
    res1 = qpc._solvers.AlphaDivergenceCvxpySolver(
        alpha=2,
        mom_atol=1e-8,
        solve_kwargs={"solver": cp.OSQP},
    )
    res1.fit(x, mu)

    res2 = qpc._solvers.AlphaDivergenceCvxpySolver(
        alpha=2,
        # mom_atol=.001, mom_rtol=.001,
        solve_kwargs={"solver": cp.OSQP},
    )
    res2.fit(x, mu)

    l2 = res2._dual_moments()
    l1 = res1._dual_moments()

    assert np.allclose(l1["dual"], l2["dual"], rtol=0.1, atol=0.001)

    assert np.allclose(res1.dual_sum1(), res2.dual_sum1(), rtol=1e-2, atol=1e-4)

    assert np.allclose(res1.dual_nonneg(), res2.dual_nonneg(), rtol=1e-2, atol=5e-4)


def test_example_norm():
    w = np.loadtxt(os.path.join(dirname, "test_example_w_norm.txt"))

    x = np.loadtxt(os.path.join(dirname, "test_example_x.txt"))

    mu = np.loadtxt(os.path.join(dirname, "test_example_mu.txt"))

    # w2 = qpc.estimate_weights_multisample(
    #    x, mu, use_norm=True, solve_kwargs={"solver": cp.ECOS}
    # )
    res = qpc._solvers.AlphaDivergenceCvxpySolver(
        alpha=2,
        use_norm=True,
        solve_kwargs={"solver": cp.ECOS},
    )
    res.fit(x, mu)
    w2 = res.weights(renormalize=True)

    assert np.allclose(w, w2)
    # assert np.allclose(w, w2, atol=1e-6, rtol=1e-4)


def test_example_alpha3():
    w = np.loadtxt(os.path.join(dirname, "test_example_w_alpha3.txt"))

    x = np.loadtxt(os.path.join(dirname, "test_example_x.txt"))

    mu = np.loadtxt(os.path.join(dirname, "test_example_mu.txt"))

    w2 = qpc.estimate_weights_multisample(
        x, mu, alpha=3, solve_kwargs={"solver": cp.ECOS}
    )

    assert np.allclose(w, w2)
    # assert np.allclose(w, w2, atol=1e-6, rtol=1e-3)


def test_example_kl():
    w = np.loadtxt(os.path.join(dirname, "test_example_w_kl.txt"))

    x = np.loadtxt(os.path.join(dirname, "test_example_x.txt"))

    mu = np.loadtxt(os.path.join(dirname, "test_example_mu.txt"))

    w2 = qpc.estimate_weights_multisample(
        x, mu, alpha="kl", use_dual_lbfgs=False, solve_kwargs={"solver": cp.ECOS}
    )

    assert np.allclose(w, w2)


def test_maxent_dual_lbfgs_eq():
    x = np.loadtxt(os.path.join(dirname, "test_example_x.txt"))

    mu = np.loadtxt(os.path.join(dirname, "test_example_mu.txt"))

    w = qpc.estimate_weights_multisample(
        x,
        mu,
        alpha="kl",
        use_dual_lbfgs=False,
        solve_kwargs={"solver": cp.ECOS},
        return_with_solver_info=True,
    )

    w2 = qpc.estimate_weights_multisample(
        x,
        mu,
        alpha="kl",
        # use_dual_lbfgs=True,
        return_with_solver_info=True,
    )

    assert np.allclose(w["weights"], w2["weights"], atol=1e-4)
    assert np.allclose(w["objective"], w2["objective"], atol=1e-3)
    assert np.allclose(w["dual"], w2["dual"], atol=1e-2, rtol=1e-2)


def test_maxent_dual_lbfgs_ineq():
    x = np.loadtxt(os.path.join(dirname, "test_example_x.txt"))

    mu = np.loadtxt(os.path.join(dirname, "test_example_mu.txt"))

    w = qpc.estimate_weights_multisample(
        x,
        mu,
        alpha="kl",
        mom_atol=0.01,
        use_dual_lbfgs=False,
        solve_kwargs={"solver": cp.ECOS},
        return_with_solver_info=True,
    )

    w2 = qpc.estimate_weights_multisample(
        x,
        mu,
        alpha="kl",
        mom_atol=0.01,
        # use_dual_lbfgs=True,
        return_with_solver_info=True,
    )

    assert np.allclose(w["weights"], w2["weights"], atol=1e-4)
    assert np.allclose(w["objective"], w2["objective"], atol=1e-3)
    assert np.allclose(w["dual"], w2["dual"], atol=1e-2, rtol=1e-2)
    assert np.allclose(w["dual_lower"], w2["dual_lower"], atol=1e-2, rtol=1e-2)
    assert np.allclose(w["dual_upper"], w2["dual_upper"], atol=1e-2, rtol=1e-2)



import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from ._quipcell import (
    estimate_size_factors,
    estimate_weights_multisample,
    renormalize_weights,
)

__all__ = [
    "estimate_size_factors",
    "estimate_weights_multisample",
    "renormalize_weights",
]



#!/usr/bin/env python

import cvxpy as cp
import numpy as np

import quipcell as qpc

mu1 = np.array([1, 0])
mu2 = np.array([0, 1])
sigma = 0.1

mu = np.array([[0.75, 0.25], [0.5, 0.5], [0.1, 0.9]])

assert np.allclose(mu.sum(axis=1), 1)

rng = np.random.default_rng(12345)

n = 10

x1 = mu1 + sigma * rng.normal(size=(n, 2))
x2 = mu2 + sigma * rng.normal(size=(n, 2))

x = np.vstack([x1, x2])

# Explicitly specify solver b/c cvxpy may change default solvers,
# e.g. in 1.5 clarabel will become default for many problems

w = qpc.estimate_weights_multisample(x, mu, solve_kwargs={"solver": cp.OSQP})

w_relax = qpc.estimate_weights_multisample(
    x, mu, mom_atol=0.001, solve_kwargs={"solver": cp.OSQP}
)

w_relax2 = qpc.estimate_weights_multisample(
    x, mu, mom_rtol=0.1, solve_kwargs={"solver": cp.OSQP}
)

w_norm = qpc.estimate_weights_multisample(
    x, mu, use_norm=True, solve_kwargs={"solver": cp.ECOS}
)

w3 = qpc.estimate_weights_multisample(x, mu, alpha=3, solve_kwargs={"solver": cp.ECOS})

w_kl = qpc.estimate_weights_multisample(
    x, mu, alpha="kl", solve_kwargs={"solver": cp.ECOS}
)


def check_weights_reasonable(w):
    assert np.allclose(w.sum(axis=0), 1)

    assert np.all(n * w[:n, 0] < 0.85) and np.all(n * w[:n, 0] > 0.65)
    assert np.all(n * w[n:, 0] < 0.35) and np.all(n * w[n:, 0] > 0.15)

    assert np.all(n * w[:, 1] > 0.4) and np.all(n * w[:, 1] < 0.6)

    assert np.all(n * w[:n, 2] < 0.25) and np.all(n * w[n:, 2] > 0.8)


check_weights_reasonable(w)
check_weights_reasonable(w_relax)
check_weights_reasonable(w_norm)

# check_weights_reasonable(w_relax2)
assert np.max(np.abs(w_relax2 - w)) < 0.01

# check_weights_reasonable(w3)
assert np.max(np.abs(w3 - w)) < 0.01

# check_weights_reasonable(w_kl)
assert np.max(np.abs(w_kl - w)) < 0.012

np.savetxt("test_example_mu.txt", mu)
np.savetxt("test_example_x.txt", x)
np.savetxt("test_example_w.txt", w)
np.savetxt("test_example_w_relaxed.txt", w_relax)
np.savetxt("test_example_w_relaxed2.txt", w_relax2)
np.savetxt("test_example_w_norm.txt", w_norm)
np.savetxt("test_example_w_alpha3.txt", w3)
np.savetxt("test_example_w_kl.txt", w_kl)
